
# Transform the h s or v coordinates of a colour
colourtransform <- function(col,h=NULL,s=NULL,v=NULL) {
  x <- rgb2hsv(col2rgb(col))
  if (!is.null(h)) x['h',] <- if (is.numeric(h)) h else h(x['h',])
  if (!is.null(s)) x['s',] <- if (is.numeric(s)) s else s(x['s',])
  if (!is.null(v)) x['v',] <- if (is.numeric(v)) v else v(x['v',])
  hsv(x['h',],x['s',],x['v',])
}

# Given a sample x1,x2,...,xn, reduce the number of points to <=len.
# Split into equal-size groups, and within each group use the FUN function
# to aggregate.
subsample <- function(x,len,FUN=function(x) x[length(x)])
  if (length(x)<=len) return(x) else
sapply(split(x,c(0,rep(1:len,each=ceiling(length(x)/len),length.out=length(x)-1))),FUN)


panel.queue <- function(t,val1,val2,val3,ev,id,type,...,groups=NULL,
                        maxpoints.queue=2000,queuescale=1,
                        col=if(is.null(groups)) plot.line$col else superpose.line$col,
                        lty=if(is.null(groups)) plot.line$lty else superpose.line$lty) {
  # Work out the groups and levels
  Groups <- if (is.null(groups)) rep(1,length.out=length(t)) else groups
  Groups <- if (is.factor(Groups)) Groups else factor(Groups)
  ngroups <- as.numeric(Groups)
  solitary <- nrow(unique(cbind(id,Groups)))==1
  #
  # Work out the colours
  plot.line <- trellis.par.get('plot.line')
  superpose.line <- trellis.par.get('superpose.line')
  col <- rep(col, length.out=nlevels(Groups))
  lightcol <- colourtransform(col,s=function(x) pmin(.8*x,.2),v=function(x) pmax(.85,1-(1-x)/2))
  fillcol <- colourtransform(col,s=function(x) pmin(.6*x,.1),v=function(x) pmax(.9,1-(1-x)/2))
  #
  for (gr in unique(ngroups)) {
    for (i in as.character(unique(id[ngroups==gr]))) {
      s <- ngroups==gr & id==i
      s <- list(t=t[s],val1=val1[s])
      d <- ngroups==gr & id==i & ev=='pkt_drop'
      d <- list(t=t[d],val1=val1[d])
      reduced <- FALSE
      if (!is.null(maxpoints.queue) && length(s$t)>maxpoints.queue) {
        reduced <- TRUE
        s <- list(t=subsample(s$t,maxpoints.queue),val1=subsample(s$val1,maxpoints.queue),
                  val2=subsample(s$val1,maxpoints.queue,FUN=min),val3=subsample(s$val1,maxpoints.queue,FUN=max))
        d <- list(t=subsample(d$t,maxpoints.queue),val1=subsample(d$val1,maxpoints.queue))
      }
      if (solitary)
        grid.polygon(x = c(s$t[1],s$t,s$t[length(s$t)]),
                     y = c(0,s$val1/queuescale,0),
                     gp = gpar(fill=fillcol[gr],lty=0), default.units='native')
      if (reduced)
        grid.polygon(x = rep(c(s$t,rev(s$t)[-1]),each=2)[-c(1,4*length(s$t)-2)],
                     y = rep(c(s$val2[-1],rev(s$val3[-1])),each=2)/queuescale,
                     gp = gpar(fill=lightcol[gr], lty=0), default.units='native')
      llines(s$t,s$val1/queuescale, col=col[gr], ...)
      lpoints(d$t,d$val1/queuescale,col='red',pch='3')
    }}
}


prepanel.queue <- function(t,val1,val2,val3,ev,id,type,...,queuescale=1)
  list(xlim=range(t),ylim=range(val1)/queuescale)

  
# Plot queue size, showing actual queue size as well as min+max in each time interval
# queuescale=NULL: no rescaling. queuescale=x: rescale queuesize to be in [0,x]
#
panel.queueapprox <- function(t,val1,val2,val3,ev,id,type,...,groups=NULL,
                            maxpoints.queueapprox=500,queuescale=NULL,queuerange=NULL,
                            col=if(is.null(groups)) plot.line$col else superpose.line$col,
                            lty=if(is.null(groups)) plot.line$lty else superpose.line$lty) {
  # Work out the groups and levels
  Groups <- if (is.null(groups)) rep(1,length.out=length(t)) else groups
  Groups <- if (is.factor(Groups)) Groups else factor(Groups)
  ngroups <- as.numeric(Groups)
  solitary <- nrow(unique(cbind(id,Groups)))==1
  #
  # Work out the colours
  plot.line <- trellis.par.get('plot.line')
  superpose.line <- trellis.par.get('superpose.line')
  col <- rep(col, length.out=nlevels(Groups))
  lightcol <- colourtransform(col,s=function(x) pmin(.8*x,.2),v=function(x) pmax(.85,1-(1-x)/2))
  fillcol <- colourtransform(col,s=function(x) pmin(.6*x,.1),v=function(x) pmax(.9,1-(1-x)/2))
  #
  buffersize <- unique(data.frame(id=id[ev=='queue_overflow'],val3=val3[ev=='queue_overflow']))
  buffersize <- structure(buffersize$val3,names=as.character(buffersize$id))
  if (!is.null(queuescale) && !is.null(queuerange)) stop("Specify either queuescale or queuerange, not both")
  if (is.null(queuescale) && is.null(queuerange)) queuescale <- 1
  qscale <- function(i) {
    if (is.null(queuerange)) return(queuescale)
    bf <- if(length(buffersize)==1) buffersize else buffersize[i]
    if (length(queuerange)==1) return(bf/queuerange)
    if (i %in% names(queuerange)) return(bf/queuerange[i])
    bf/queuerange[1] }
  for (gr in unique(ngroups)) {
    for (i in as.character(unique(id[ngroups==gr]))) {
      s <- ev=='queue_range' & ngroups==gr & id==i
      s <- list(t=t[s],val1=val1[s],val2=val2[s],val3=val3[s])
      if (!is.null(maxpoints.queueapprox))
        s <- list(t=subsample(s$t,maxpoints.queueapprox),val1=subsample(s$val1,maxpoints.queueapprox),
                  val2=subsample(s$val2,maxpoints.queueapprox,FUN=min),val3=subsample(s$val3,maxpoints.queueapprox,FUN=max))
      if (solitary)
        grid.polygon(x = c(s$t[1],s$t,s$t[length(s$t)]),
                     y = c(0,s$val1/qscale(i),0),
                     gp = gpar(fill=fillcol[gr],lty=0), default.units='native')
      grid.polygon(x = rep(c(s$t,rev(s$t)[-1]),each=2)[-c(1,4*length(s$t)-2)],
                   y = rep(c(s$val2[-1],rev(s$val3[-1])),each=2)/qscale(i),
                   gp = gpar(fill=lightcol[gr], lty=0), default.units='native')
      }}
  for (gr in unique(ngroups)) {
    for (i in as.character(unique(id[ngroups==gr]))) {
      s <- ev=='queue_range' & ngroups==gr & id==i
      s <- list(t=t[s],val1=val1[s],val2=val2[s],val3=val3[s])
      if (!is.null(maxpoints.queueapprox))
        s <- list(t=subsample(s$t,maxpoints.queueapprox),val1=subsample(s$val1,maxpoints.queueapprox),
                  val2=subsample(s$val2,maxpoints.queueapprox,FUN=min),val3=subsample(s$val3,maxpoints.queueapprox,FUN=max))
      llines(s$t,s$val1/qscale(i), col=col[gr], ...)
    }}
}

prepanel.queueapprox <- function(t,val1,val2,val3,ev,id,type,...,groups=NULL,
                               maxpoints.queueapprox=500,queuescale=1,queuerange=NULL) {
  # Work out the groups and levels
  Groups <- if (is.null(groups)) rep(1,length.out=length(t)) else groups
  Groups <- if (is.factor(Groups)) Groups else factor(Groups)
  ngroups <- as.numeric(Groups)
  solitary <- nrow(unique(cbind(id,Groups)))==1
  #
  ylim <- c(Inf,-Inf)
  #
  buffersize <- unique(data.frame(id=id[ev=='queue_overflow'],val3=val3[ev=='queue_overflow']))
  buffersize <- structure(buffersize$val3,names=as.character(buffersize$id))
  if (!is.null(queuescale) && !is.null(queuerange)) stop("Specify either queuescale or queuerange, not both")
  if (is.null(queuescale) && is.null(queuerange)) queuescale <- 1
  qscale <- function(i) {
    if (is.null(queuerange)) return(queuescale)
    bf <- if(length(buffersize)==1) buffersize else buffersize[i]
    if (length(queuerange)==1) return(bf/queuerange)
    if (i %in% names(queuerange)) return(bf/queuerange[i])
    bf/queuerange[1] }
  for (gr in unique(ngroups)) {
    for (i in as.character(unique(id[ngroups==gr]))) {
      s <- ev=='queue_range' & ngroups==gr & id==i
      s <- list(t=t[s],val1=val1[s],val2=val2[s],val3=val3[s])
      if (!is.null(maxpoints.queueapprox))
        s <- list(t=subsample(s$t,maxpoints.queueapprox),val1=subsample(s$val1,maxpoints.queueapprox),
                  val2=subsample(s$val2,maxpoints.queueapprox,FUN=min),val3=subsample(s$val3,maxpoints.queueapprox,FUN=max))
      ylim <- c(min(ylim[1],min(s$val2)/qscale(i)),max(ylim[2],max(s$val3)/qscale(i)))
    }}
  list(xlim=range(t),ylim=ylim)
}


# Plot utilization and drops.
# Utilization is found by measuring the amount of idle time.
# Drops are found by measuring the "local time at queue=full".
#
panel.queuerecord <- function(t,val1,val2,val3,ev,id,type,...,groups=NULL,
                       maxpoints.queuerecord=500,type.load=c('drops','utilization'),y=list(drops=1,utilization=0),
                       col=if(is.null(groups)) plot.line$col else superpose.line$col,
                       lty=if(is.null(groups)) plot.line$lty else superpose.line$lty) {
  # Work out the groups and levels
  Groups <- if (is.null(groups)) rep(1,length.out=length(t)) else groups
  Groups <- if (is.factor(Groups)) Groups else factor(Groups)
  ngroups <- as.numeric(Groups)
  solitary <- nrow(unique(cbind(id,Groups)))==1
  #
  # Work out the colours
  plot.line <- trellis.par.get('plot.line')
  superpose.line <- trellis.par.get('superpose.line')
  col <- rep(col, length.out=nlevels(Groups))
  lightcol <- colourtransform(col,s=function(x) pmin(.8*x,.2),v=function(x) pmax(.85,1-(1-x)/2))
  fillcol <- colourtransform(col,s=function(x) pmin(.6*x,.1),v=function(x) pmax(.9,1-(1-x)/2))
  redcol <- colourtransform(col,h=0,s=.5,v=function(x) pmax(.95,1-(1-x)/2))
  #
  for (gr in unique(ngroups)) {
    for (i in as.character(unique(id[ngroups==gr]))) {
      s <- ev=='cum_traffic' & ngroups==gr & id==i
      s <- list(t=t[s],val1=val1[s],val2=val2[s],val3=val3[s])
      if (!is.null(maxpoints.queuerecord))
        s <- list(t=subsample(s$t,maxpoints.queuerecord),val1=subsample(s$val1,maxpoints.queuerecord),
                  val2=subsample(s$val2,maxpoints.queuerecord),val3=subsample(s$val3,maxpoints.queuerecord))
      if ('drops' %in% type.load) {
        drops <- diff(s$val3)/diff(s$t)
        if (solitary)
          grid.polygon(x = rep(s$t,each=2),
                       y = c(0,rep(drops,each=2),0)+y$drops,
                       gp = gpar(fill=redcol[gr],lty=0), default.units='native')
        drops <- c(drops[1],drops)
        w <- drops>0 | c(0,drops[-length(drops)])>0 | c(drops[-1],0)>0
        llines(s$t,ifelse(w,drops,NA)+y$drops,type='S',col=col[gr])
      }
      if ('utilization' %in% type.load) {
        idle <- diff(s$val2)/diff(s$t)
        if (solitary)
          grid.polygon(x = rep(s$t,each=2),
                       y = c(0,rep(1-idle,each=2),0)+y$utilization,
                       gp = gpar(fill=fillcol[gr],lty=0), default.units='native')
        llines(s$t,y$utilization+1-c(idle[1],idle),type='S',col=col[gr])
      }
  }}
}

prepanel.queuerecord <- function(t,val1,val2,val3,ev,id,type,...,groups=NULL,
                       maxpoints.queuerecord=500,type.load=c('drops','utilization'),y=list(drops=1,utilization=0)) {
  # Work out the groups and levels
  Groups <- if (is.null(groups)) rep(1,length.out=length(t)) else groups
  Groups <- if (is.factor(Groups)) Groups else factor(Groups)
  ngroups <- as.numeric(Groups)
  solitary <- nrow(unique(cbind(id,Groups)))==1
  #
  ylim <- c(Inf,-Inf)
  #
  for (gr in unique(ngroups)) {
    for (i in as.character(unique(id[ngroups==gr]))) {
      s <- ev=='cum_traffic' & ngroups==gr & id==i
      s <- list(t=t[s],val1=val1[s],val2=val2[s],val3=val3[s])
      if (!is.null(maxpoints.queuerecord))
        s <- list(t=subsample(s$t,maxpoints.queuerecord),val1=subsample(s$val1,maxpoints.queuerecord),
                  val2=subsample(s$val2,maxpoints.queuerecord),val3=subsample(s$val3,maxpoints.queuerecord))
      if ('drops' %in% type.load) {
        drops <- diff(s$val3)/diff(s$t)
        ylim <- c(min(ylim[1],min(drops)+y$drops),max(ylim[2],max(drops)+y$drops))
      }
      if ('utilization' %in% type.load) {
        idle <- diff(s$val2)/diff(s$t)
        ylim <- c(min(ylim[1],min(1-idle)+y$utilization),max(ylim[2],max(1-idle)+y$utilization))
      }
  }}
  list(xlim=range(t),ylim=round(ylim,2))
}

    
# panel.queue does panel.queuesize (queuescale=scaled to be in [0,1]) and panel.queuerecord.
# Or, it does just panel.queuesize (queuescale=NULL).
# Or, it does just panel.queuerecord.
# It chooses, based on the type of data it is given.
#
panel.queue2 <- function(t,val1,val2,val3,ev,id,type,...,groups=NULL,
                        queuescale=NULL,queuerange=NULL) {
  tt <- sort(as.character(unique(type)))
  if (identical(tt,'queueapprox'))
    panel.queueapprox(t,val1,val2,val3,ev,id,type,...,groups=groups,queuescale=queuescale,queuerange=queuerange)
  else if (identical(tt,'queuerecord'))
    panel.queuerecord(t,val1,val2,val3,ev,id,type,...,groups=groups)
  else if (identical(tt,c('queueapprox','queuerecord'))) {
    ref <- trellis.par.get('reference.line')
    panel.abline(h=c(0,1),col=ref$col,lty=ref$lty,lwd=ref$lwd)
    panel.queueapprox(t,val1,val2,val3,ev,id,type,...,groups=groups,
                    queuescale=NULL,queuerange=if(is.null(queuerange)) 1 else queuerange)
    panel.queuerecord(t,val1,val2,val3,ev,id,type,...,groups=groups,y=list(drops=1,utilization=-1))
  }
  else stop("Don't know how to plot a queue panel with type=",tt)
}
prepanel.queue2 <- function(t,val1,val2,val3,ev,id,type,...,
                           queuescale=NULL,queuerange=NULL) {
  tt <- sort(as.character(unique(type)))
  if (identical(tt,'queueapprox'))
    prepanel.queueapprox(t,val1,val2,val3,ev,id,type,...,queuescale=queuescale,queuerange=queuerange)
  else if (identical(tt,'queuerecord'))
    prepanel.queuerecord(t,val1,val2,val3,ev,id,type,...)
  else if (identical(tt,c('queueapprox','queuerecord'))) {
    r1 <- prepanel.queueapprox(t,val1,val2,val3,ev,id,type,...,
                             queuescale=NULL,queuerange=if(is.null(queuerange)) 1 else queuerange)
    r2 <- prepanel.queuerecord(t,val1,val2,val3,ev,id,type,...,y=list(drops=1,utilization=-1))
    list(xlim=range(t),ylim=c(min(r1$ylim[1],r2$ylim[1]),max(r1$ylim[2],r2$ylim[2])))
  }
  else stop("Don't know how to plot a queue panel with type=",tt)
}


tcpstyles.default <- list(col=c('tcp_rcv'=NA,'tcp_rcv_dup'='orange','tcp_rcv_dup_fastxmit'='red',
                            'tcp_rcv_dup_fr'='orange','tcp_rcv_dup_fr_end'='green',
                            'tcp_rcv_fr'='purple','tcp_rcv_3dupnofr'='black','tcp_timeout'='black'),
                          pch=c('tcp_rcv'=NA,'tcp_rcv_dup'=0,'tcp_rcv_dup_fastxmit'=3,
                            'tcp_rcv_dup_fr'=1,'tcp_rcv_dup_fr_end'=16,
                            'tcp_rcv_fr'=8,'tcp_rcv_3dupnofr'=5,'tcp_timeout'=18))
tcpstyle <- function(ev,default=plot.symbol) {
  plot.symbol <- trellis.par.get('plot.symbol')
  col <- ifelse(is.na(tcpstyles.default$col),default$col,tcpstyles.default$col)
  pch <- ifelse(is.na(tcpstyles.default$pch),default$pch,tcpstyles.default$pch)
  list(col=col[as.character(ev)],pch=pch[as.character(ev)])
}


# Plot one or several TCP flows.
# If it's just one flow (or illustrate=TRUE), then use a detailed key
panel.tcp <- function(t,val1,val2,val3,ev,id,type,...,tcpscale=1,
                      cex=.6,tcpsymbols='b',groups=NULL,illustrate=NULL,
                      col=if(is.null(groups)) plot.line$col else superpose.line$col,
                      lty=if(is.null(groups)) plot.line$lty else superpose.line$lty,
                      pch=if(is.null(groups)) plot.symbol$pch else superpose.symbol$pch,
                      tcpkey=tcpstyle) {
  # Work out the groups and levels
  Groups <- if (is.null(groups)) rep(1,length.out=length(t)) else groups
  Groups <- if (is.factor(Groups)) Groups else factor(Groups)
  ngroups <- as.numeric(Groups)
  solitary <- if (is.null(illustrate)) nrow(unique(cbind(id,Groups)))==1 else illustrate
  #
  # Work out the colours
  plot.line <- trellis.par.get('plot.line')
  superpose.line <- trellis.par.get('superpose.line')
  plot.symbol <- trellis.par.get('plot.symbol')
  superpose.symbol <- trellis.par.get('superpose.symbol')
  col <- rep(col, length.out=nlevels(Groups))
  lty <- rep(lty, length.out=nlevels(Groups))
  pch <- rep(pch, length.out=nlevels(Groups))
  #
  for (gr in unique(ngroups)) {
    for (i in as.character(unique(id[ngroups==gr]))) {
      s <- type=='tcp' & ngroups==gr & id==i
      s <- list(t=t[s],val1=val1[s],val2=val2[s],val3=val3[s],ev=ev[s])
      if (solitary) {
        llines(s$t,s$val3/tcpscale,...,type='l',col=col[gr],lty=lty[gr],cex=cex)
        colpch <- tcpkey(s$ev,list(col=col[gr],pch=pch[gr]))
        lpoints(s$t,s$val3/tcpscale,...,type='p',col=colpch$col,pch=colpch$pch,cex=cex)
      }
      else
        llines(s$t,s$val3/tcpscale,...,type=tcpsymbols,col=col[gr],lty=lty[gr],cex=cex)
    }}
}
prepanel.tcp <- function(t,val1,val2,val3,ev,id,type,...,tcpscale=1,groups=NULL) {
  list(xlim=range(t),ylim=range(c(0,val3[type=='tcp']))/tcpscale*1.05)
}




# Helper function, to parse something like .~time|id, data=res
#
traceplotParseFormula <- function (model, data, subset=TRUE, groups=NULL, subscripts=TRUE, drop=NULL) {
  expr2char <- function(x) paste(deparse(x), collapse = "")
  if (inherits(groups, "formula")) {
    groupVar <- as.character(groups)[2]
    groups <- eval(parse(text = groupVar), data, environment(groups))
  }
  parseSide <- function(model) {
    model.vars <- list()
    while (length(model) == 3 && model[[1]] == as.name("+")) {
      model.vars <- c(model.vars, model[[3]])
      model <- model[[2]]
    }
    rev(c(model.vars, model))
  }
  parseCond <- function(model) {
    model <- substitute(~m, list(m = model))[[2]]
    model.vars <- list()
    while (length(model) == 3 && (model[[1]] == as.name("*") || 
                   model[[1]] == as.name("+"))) {
      model.vars <- c(model.vars, model[[3]])
      model <- model[[2]]
    }
    rev(c(model.vars, model))
  }
  concat <- function(arglist) {
    if (length(arglist)==1) return(arglist[[1]])
    if (any(sapply(arglist, is.factor))) return(factor(unlist(lapply(arglist, as.character))))
    if (any(sapply(arglist, is.shingle))) stop("shingles can not be concatenated")
    do.call("c", arglist)
  }
  #
  # Work out the two sides to the equation
  if (!inherits(model, "formula")) stop("model must be a formula object")
  ans <- list(left=NULL, right=NULL, condition=NULL, left.name=character(0), right.name=character(0))
  if (length(model) != 3) stop("Formula needs both a LHS and a RHS")
  # LHS
  varsLHS <- if (!identical(model[[2]],as.name('.'))) parseSide(model[[2]]) else parseSide(as.formula(val1+val2+val3+ev+id+type~.)[[2]])
  if (length(varsLHS)!=6) stop("LHS must be . or of the form val1+val2+val3+ev+id+type")
  # RHS
  modelRHS <- model[[length(model)]]
  if (length(modelRHS) == 3 && modelRHS[[1]] == as.name("|")) {
    modelRHS.vars <- parseCond(modelRHS[[3]])
    modelRHS <- modelRHS[[2]]
    varsRHS <- list(modelRHS)
    if (length(varsRHS)!=1) stop("Need a single RHS variable")
  }
  else {
    varsRHS <- list(modelRHS)
    modelRHS.vars <- list()
  }
  #
  # Decide what to drop
  if (is.null(drop)) drop <- TRUE
  if (is.list(drop)) {
    drop.unused.cond <- if (is.null(drop$cond)) TRUE else drop$cond
    drop.unused.data <- if (is.null(drop$data)) TRUE else drop$data
  }
  else {
    drop.unused.cond <- drop
    drop.unused.data <- drop
  }
  # Evaluate the two sides
  env <- environment(model)
  # LHS
  ans$left.name <- 'state'
  ans$left <- lapply(varsLHS, function(i) {
    tmp <- eval(i, data, env)
    if (!is.matrix(tmp)) tmp <- if (is.factor(tmp) || is.shingle(tmp)) tmp[subset, drop = drop.unused.data] else tmp[subset]
    if (inherits(tmp, "POSIXt")) tmp <- as.POSIXct(tmp)
    tmp })
  names(ans$left) <- c('val1','val2','val3','ev','id','type')
  # RHS
  tmp <- eval(varsRHS[[1]], data, env)
  if (is.matrix(tmp)) tmp <- as.data.frame(tmp)
  nobs <- if (is.data.frame(tmp)) nrow(tmp) else length(tmp)
  ans$right <- (if (is.data.frame(tmp)) tmp[subset,] else
                (if (is.factor(tmp) || is.shingle(tmp)) tmp[subset,drop=drop.unused.data] else tmp[subset]))
  ans$right.name <- expr2char(modelRHS)
  # Cond
  if (length(modelRHS.vars)>0) {
    ans$condition <- vector("list", length(modelRHS.vars))
    names(ans$condition) <- sapply(modelRHS.vars, expr2char)
    for (i in seq_along(modelRHS.vars))
      ans$condition[[i]] <- as.factorOrShingle(eval(modelRHS.vars[[i]],data,env),subset, drop=drop.unused.cond)
  }
  #
  # Work out groups
  if (!is.null(groups)) ans$groups <- groups
  if (subscripts) ans$subscr <- seq_len(nobs)[subset]
  #
  ans
}



traceplot <- function(x, data = NULL,
                      auto.key = FALSE, aspect = "fill", 
                      panel = panel.traceplot, prepanel = NULL, 
                      scales = list(), strip = TRUE, groups = NULL, xlab, xlim, 
                      ylab, ylim, drop.unused.levels = lattice.getOption("drop.unused.levels"), 
                      ..., lattice.options = NULL, default.scales = list(), subscripts = !is.null(groups), 
                      subset = TRUE) {
  updateList <- getFromNamespace('updateList','lattice')
  construct.scales <- getFromNamespace('construct.scales','lattice')
  compute.packet <- getFromNamespace('compute.packet','lattice')
  cupdate <- getFromNamespace('cupdate','lattice')
  limits.and.aspect <- getFromNamespace('limits.and.aspect','lattice')
  cond.orders <- getFromNamespace('cond.orders','lattice')
  #  
  formula <- x
  dots <- list(...)
  groups <- eval(substitute(groups), data, environment(x))
  subset <- eval(substitute(subset), data, environment(x))
  if (!is.null(lattice.options)) {
    oopt <- lattice.options(lattice.options)
    on.exit(lattice.options(oopt), add = TRUE)
  }
  form <- traceplotParseFormula(formula, data, subset=subset, groups=groups, subscripts=TRUE, drop=drop.unused.levels)
  groups <- form$groups
  if (!is.function(panel)) panel <- eval(panel)
  if (!is.function(strip)) strip <- eval(strip)
  subscr <- form$subscr
  prepanel <- if (is.function(prepanel)) prepanel else (if (is.character(prepanel)) get(prepanel) else eval(prepanel))
  cond <- form$condition
  y <- form$left
  x <- form$right
  if (length(cond) == 0) {
    strip <- FALSE
    cond <- list(gl(1, length(x)))
  }
  if (missing(xlab)) xlab <- form$right.name
  if (missing(ylab)) ylab <- form$left.name
  f <- getFromNamespace('trellis.skeleton','lattice')
  foo <- do.call(f, c(list(formula = formula, 
                           cond = cond, aspect = aspect, strip = strip, panel = panel, 
                           xlab = xlab, ylab = ylab, xlab.default = form$right.name, 
                           ylab.default = form$left.name, lattice.options = lattice.options), 
                      dots))
  dots <- foo$dots
  foo <- foo$foo
  foo$call <- sys.call(sys.parent())
  foo$call[[1]] <- quote(xyplot)
  if (is.character(scales)) scales <- list(relation = scales)
  scales <- updateList(default.scales, scales)
  foo <- c(foo, do.call("construct.scales", scales))
  have.xlim <- !missing(xlim)
  if (!is.null(foo$x.scales$limit)) {
    have.xlim <- TRUE
    xlim <- foo$x.scales$limit
  }
  have.ylim <- !missing(ylim)
  if (!is.null(foo$y.scales$limit)) {
    have.ylim <- TRUE
    ylim <- foo$y.scales$limit
  }
  cond.max.level <- unlist(lapply(cond, nlevels))
  foo$panel.args.common <- dots
  foo$panel.args.common$groups <- groups
  npackets <- prod(cond.max.level)
  if (npackets != prod(sapply(foo$condlevels, length))) 
    stop("mismatch in number of packets")
  foo$panel.args <- vector(mode = "list", length = npackets)
  foo$packet.sizes <- numeric(npackets)
  if (npackets > 1) {
    dim(foo$packet.sizes) <- sapply(foo$condlevels, length)
    dimnames(foo$packet.sizes) <- lapply(foo$condlevels, as.character)
  }
  cond.current.level <- rep(1, length(cond))
  for (packet.number in seq_len(npackets)) {
    id <- compute.packet(cond, cond.current.level)
    foo$packet.sizes[packet.number] <- sum(id)
    foo$panel.args[[packet.number]] <- list(x = x[id], y = lapply(y, function(yy) yy[id]))
    foo$panel.args[[packet.number]]$subscripts <- subscr[id]
    cond.current.level <- cupdate(cond.current.level, cond.max.level)
  }
  more.comp <- c(limits.and.aspect(prepanel.default.traceplot, 
                                   prepanel = prepanel, have.xlim = have.xlim, xlim = xlim, 
                                   have.ylim = have.ylim, ylim = ylim, x.relation = foo$x.scales$relation, 
                                   y.relation = foo$y.scales$relation, panel.args.common = foo$panel.args.common, 
                                   panel.args = foo$panel.args, aspect = aspect, npackets = npackets, 
                                   x.axs = foo$x.scales$axs, y.axs = foo$y.scales$axs), 
                 cond.orders(foo))
  foo[names(more.comp)] <- more.comp
  if (is.null(foo$legend) && !is.null(groups) && (is.list(auto.key) || 
                                                  (is.logical(auto.key) && auto.key))) {
    foo$legend <- list(list(fun = "drawSimpleKey", args = updateList(list(text = levels(as.factor(groups)), 
                                                     points = TRUE, rectangles = FALSE, lines = FALSE), 
                                                     if (is.list(auto.key)) auto.key else list())))
    foo$legend[[1]]$x <- foo$legend[[1]]$args$x
    foo$legend[[1]]$y <- foo$legend[[1]]$args$y
    foo$legend[[1]]$corner <- foo$legend[[1]]$args$corner
    names(foo$legend) <- if (any(c("x", "y", "corner") %in% names(foo$legend[[1]]$args))) "inside" else "top"
    if (!is.null(foo$legend[[1]]$args$space)) 
      names(foo$legend) <- foo$legend[[1]]$args$space
  }
  class(foo) <- "trellis"
  foo
}



panel.traceplot <- function(x,y,...,groups,subscripts) {
  tt <- sort(as.character(unique(y$type)))
  if (identical(tt,c('queueapprox','queuerecord'))) tt <- 'queue2'
  if (length(tt)==1) {
    args <- c(list(x),y[c('val1','val2','val3','ev','id','type')],list(...),list(subscripts=subscripts))
    if (!missing(groups)) args <- c(args,list(groups=groups[subscripts]))
    do.call(paste('panel',tt,sep='.'),args)
  }
  else {
    warning(paste('Panel of mixed type ',paste(tt,collapse='+'),sep=''),call.=FALSE)
    panel.xyplot(x,y$val1,...,groups=groups,subscripts=subscripts)
  }
}
prepanel.default.traceplot <- function(x,y,...,groups,subscripts) {
  tt <- sort(as.character(unique(y$type)))
  if (identical(tt,c('queueapprox','queuerecord'))) tt <- 'queue2'
  if (length(tt)==1) {
    args <- c(list(x),y[c('val1','val2','val3','ev','id','type')],list(...),list(subscripts=subscripts))
    if (!missing(groups)) args <- c(args,list(groups=groups[subscripts]))
    do.call(paste('prepanel',tt,sep='.'),args)
  }
  else
    list(xlim=range(x),ylim=range(y$val1))
}


trafficstyles.default <- list(pch=c('pkt_arrive'=3,'pkt_depart'=4,'pkt_createsend'=15,'pkt_rcvdestroy'=16,'pkt_discard'=13),
                              lty=c('pkt_arrive'=3,'pkt_depart'=3,'pkt_createsend'=1,'pkt_rcvdestroy'=1,'pkt_discard'=1),
                              col=c('pkt_arrive'=1,'pkt_depart'=0,'pkt_createsend'=1,'pkt_rcvdestroy'=0,'pkt_discard'=1),
                              cex=c('pkt_arrive'=1,'pkt_depart'=1,'pkt_createsend'=1,'pkt_rcvdestroy'=1,'pkt_discard'=2))
trafficstyle <- function(ev,default=plot.line) {
  plot.line <- trellis.par.get('plot.line')
  darkcol <- default$col[1]
  lightcol <- colourtransform(darkcol,s=function(x) pmin(.8*x,.4),v=function(x) pmax(.7,1-(1-x)*2/3))
  col <- ifelse(trafficstyles.default$col==1,darkcol,lightcol)
  pch <- trafficstyles.default$pch
  lty <- trafficstyles.default$lty
  cex <- trafficstyles.default$cex
  list(col=col[as.character(ev)],pch=pch[as.character(ev)],lty=lty[as.character(ev)],cex=cex[as.character(ev)])
}
#
panel.traffic <- function(t,val1,val2,val3,ev,id,type,...,groups=NULL,trafficscale=1,
                          cex=.6,
                          col=if(is.null(groups)) plot.line$col else superpose.line$col,
                          traffickey=trafficstyle) {
  # Work out the groups and levels
  Groups <- if (is.null(groups)) rep(1,length.out=length(t)) else groups
  Groups <- if (is.factor(Groups)) Groups else factor(Groups)
  ngroups <- as.numeric(Groups)
  #
  # Work out the colours
  plot.line <- trellis.par.get('plot.line')
  superpose.line <- trellis.par.get('superpose.line')
  col <- rep(col, length.out=nlevels(Groups))
  df <- data.frame(Gr=Groups,n=ngroups)
  #
  for (gr in unique(ngroups)) {
    for (e in as.character(unique(ev[ngroups==gr]))) {
      s <- ngroups==gr & ev==e
      s <- list(t=t[s],val1=val1[s],val2=val2[s],val3=val3[s],ev=ev[s])
      style <- traffickey(e,list(col=col[gr]))
      llines(s$t,s$val2/trafficscale,...,type='b',cex=cex*style$cex,col=style$col,pch=style$pch,lty=style$lty)
    }}
}
prepanel.traffic <- function(t,val1,val2,val3,ev,id,type,...,trafficscale=1,groups=NULL) {
  list(xlim=range(t),ylim=range(val2[type=='traffic'])/trafficscale)
}
